// [[Rcpp::depends(RcppArmadillo)]]
// [[Rcpp::interfaces(r, cpp)]]
#include <RcppArmadillo.h>
#include <math.h>
using namespace Rcpp;

double hnum(int index,int base){

	double r = 0;
	double f = 1;
	int i = index;

	while(i >0){
		f = f / base;
		r = r + f * (i % base);
		i = floor(i / base);
	}
	return(r);
}

//' Get n prime numbers
//' @param n the number of prime numbers to get, starting from 3
//' @export getPrimes
// [[Rcpp::export]]
arma::vec getPrimes(int n){

	int primegot = 0;
	int i = 3;

	int isprime = 0;

	arma::vec result(n);

	while(primegot < n){

		for(int j = 2; j < i ; j++){
			if( i % j == 0){
				isprime=1;
				break;
			}
		}

		if(isprime == 0){
			result[primegot] = i;
			primegot++;
		}

		i += 2;
		isprime = 0;
	}

	return(result);

}

//' Generate a halton sequence from a base number
//' @param H the length of the halton sequnce
//' @param prime the base number used to construct the Halton sequence, defaults to 3
//' @param burn the number of elements to "burn" from the beginning of the halton sequence, defualts to 30
//' @export hseq
// [[Rcpp::export]]
NumericVector hseq(int H, int prime = 3, int burn = 30){

	// The initial element can't be less than 1
	if( burn < 0){
		burn = 0;
	}

	NumericVector out(H);
	int j;

	for(int i = 0 ; i < H ; i++){
		j = i + burn + 1;
		out[i] = hnum(j,prime);
	}

	return(out);

}

//' Generate uniformly distributed numbers using a Halton sequence
//' @param H the length of the halton sequnce
//' @param min the lower bound of the distribution
//' @param max the upper bound of the distribution
//' @param prime the base number used to construct the Halton sequence, defaults to 3
//' @param burn the number of elements to "burn" from the beginning of the halton sequence, defualts to 30
//' @export hunif
// [[Rcpp::export]]
NumericVector hunif(int H, double min = 0, double max = 1, int prime = 3, int burn = 30){

	// The initial element can't be less than 1
	if( burn < 0){
		burn = 0;
	}
	if( prime < 1){
		prime = 3;
	}

	NumericVector out(H);
	int j;

	for(int i = 0 ; i < H ; i++){
		j = i + burn + 1;
		out[i] = hnum(j,prime);
	}

	out = out*(max - min) + min;

	return(out);

}

//' Simulate draws from a normal distribution using a halton sequence
//' @param H the length of the halton sequnce
//' @param mean the mean of the normal distribution
//' @param sd the standard deviation of the normal distribution
//' @param prime the base number used to construct the Halton sequence, defaults to 3
//' @param burn the number of elements to "burn" from the beginning of the halton sequence, defualts to 30
//' @export hnorm
// [[Rcpp::export]]
NumericVector hnorm(int H, double mean = 0.0, double sd = 1.0, int prime = 3, int burn = 30){
	NumericVector HH = hseq(H, prime, burn);
	NumericVector res = qnorm(HH, mean, sd);
	return(res);
}

//' Simulate draws from a gamma distribution using a halton sequence
//' @param H the length of the halton sequnce
//' @param shape the shape parameter of the gamma distribution
//' @param scale the scale parameter of the gamma distribution
//' @param prime the base number used to construct the Halton sequence, defaults to 3
//' @param burn the number of elements to "burn" from the beginning of the halton sequence, defualts to 30
//' @export hgamma
// [[Rcpp::export]]
NumericVector hgamma(int H, double shape = 0, double scale = 1, int prime = 3, int burn = 30){

	NumericVector HH = hseq(H, prime, burn);
	NumericVector res = qgamma(HH, shape, scale);
	return(res);
}


//' Simulate draws from a Log-Normal distribution using a halton sequence
//' @param H the length of the halton sequnce
//' @param mean the mean of the distribution
//' @param sd the standard deviation of the distribution
//' @param prime the base number used to construct the Halton sequence, defaults to 3
//' @param burn the number of elements to "burn" from the beginning of the halton sequence, defualts to 30
//' @export hlnorm
// [[Rcpp::export]]
NumericVector hlnorm(int H, double mean = 0.5, double sd = 0.25, int prime = 3, int burn = 30){

	double mu = log(mean / sqrt(1 + ( pow(sd,2) /  pow(mean,2) )));
	double std = sqrt( log(1 + ( pow(sd,2) /  pow(mean,2) )));

	NumericVector HH = hseq(H, prime, burn);
	NumericVector res = exp(qnorm(HH, mu, std));
	return(res);
}

//' Multivariate normal inversion.
//' @param H the length of the halton sequnce
//' @param means the mean of the distribution
//' @param sigma the covariance matrix
//' @export mvhnormInv
// [[Rcpp::export]]
NumericMatrix mvhnormInv( int H, arma::rowvec means, arma::mat sigma, int burn = 30){

	int ncols = sigma.n_cols;
	arma::mat Y(H, ncols);

	NumericVector hcol;

	arma::vec primes = getPrimes(ncols);

	// Generate Standard Normal Distributions
	for(int i = 0; i < ncols ; i++){
		burn += primes(i);
		hcol = hnorm(H , 0.0, 1.0, primes(i), burn);
		Y.col(i) = Rcpp::as<arma::colvec>(hcol);
	}

	arma::mat RES = arma::repmat(means, H, 1) + Y * arma::chol(sigma);

	NumericMatrix OUT(H, ncols);

	arma::colvec vtemp;
	NumericVector htemp;

	double mean;
	double sd;

	for(int i = 0; i < ncols ; i++){
		vtemp = RES.col(i);
		htemp = as<NumericVector>(wrap(vtemp));

		mean = means[i];
		sd = sqrt(sigma(i,i));

		OUT(_,i) = pnorm( htemp , mean , sd );
	}

	return(OUT);

}

//' Multivariate normal.
//' @param H the length of the halton sequnce
//' @param means the mean of the distribution
//' @param sigma the covariance matrix
//' @export mvhnorm
// [[Rcpp::export]]
arma::mat mvhnorm( int H, arma::rowvec means, arma::mat sigma, int burn = 30){

	int ncols = sigma.n_cols;
	arma::mat Y(H, ncols);

	NumericVector hcol;

	arma::vec primes = getPrimes(ncols);

	// Generate Standard Normal Distributions
	for(int i = 0; i < ncols ; i++){
		burn += primes(i);
		hcol = hnorm(H , 0.0, 1.0, primes(i), burn);
		Y.col(i) = Rcpp::as<arma::colvec>(hcol);
	}

	arma::mat RES = arma::repmat(means, H, 1) + Y * arma::chol(sigma);

	return(RES);

}


